<?php

namespace eluhr\flysystemRestApi\behaviors;

use eluhr\flysystemRestApi\events\PluginEvent;
use eluhr\flysystemRestApi\interfaces\PluginInterface;
use eluhr\flysystemRestApi\models\StorageItem;
use Yii;
use yii\base\Behavior;
use yii\base\InvalidConfigException;
use yii\db\BaseActiveRecord;

/**
 * @property StorageItem $owner
 */
class PluginsBehavior extends Behavior
{
    /**
     * @var PluginInterface[]
     */
    private array $_plugins = [];

    /**
     * @param PluginInterface[] $plugins
     *
     * @throws InvalidConfigException
     */
    public function setPlugins(array $plugins): void
    {
        foreach ($plugins as $name => $value) {

            // Ensure, that plugin is instance of [PluginInterface]]
            $plugin = $this->createPluginInstance($value);
            // If name is not set in config, use name defined in plugin
            if (is_numeric($name)) {
                $name = $plugin->name();
            }
            $this->addPlugin($name, $plugin);
        }
    }

    /**
     * @return PluginInterface[]
     */
    public function getPlugins(): array
    {
        return $this->_plugins;
    }

    /**
     * @param PluginInterface|array|string $value Either an instance of a [[PluginInterface]] or a object config
     *
     * @throws InvalidConfigException
     * @return PluginInterface
     */
    protected function createPluginInstance($value): PluginInterface {
        if (!$value instanceof PluginInterface) {
            $plugin = Yii::createObject($value);
        } else {
            $plugin = $value;
        }

        return $plugin;
    }

    /**
     * Add a plugin to the component
     *
     * @return void
     */
    public function addPlugin(string $name,PluginInterface $plugin)
    {
        $this->_plugins[$name] = $plugin;
    }

    /**
     * Get a plugin by name. If it does not exist, null will be returned
     *
     * @param string $name
     *
     * @return PluginInterface|null
     */
    public function getPlugin(string $name): ?PluginInterface
    {
        return $this->_plugins[$name] ?? null;
    }

    /**
     * Remove a plugin by name
     *
     * @param string $name
     *
     * @return void
     */
    public function removePlugin(string $name): void
    {
        unset($this->_plugins[$name]);
    }

    /**
     * @inheritdoc
     */
    public function events(): array
    {
        $events = parent::events();
        $events[BaseActiveRecord::EVENT_AFTER_FIND] = 'installPluginsInternal';
        return $events;
    }

    /**
     * Apply added plugins. Do not use directly! For internal use only! Method is public because behavior need to
     * access this method when registering events.
     *
     * @return void
     * @internal
     */
    public function installPluginsInternal()
    {
        $this->owner->trigger(PluginEvent::EVENT_BEFORE_APPLY_PLUGINS, new PluginEvent());
        foreach ($this->_plugins as $plugin) {
            $this->owner->trigger(PluginEvent::EVENT_APPLY_PLUGIN, new PluginEvent(['plugin' => $plugin]));
            $plugin->apply($this->owner);
        }
    }
}
