<?php

namespace eluhr\flysystemRestApi\components;

use eluhr\flysystemRestApi\behaviors\PluginsBehavior;
use eluhr\flysystemRestApi\helpers\StorageItemHelper;
use eluhr\flysystemRestApi\interfaces\FileStorageComponentInterface;
use eluhr\flysystemRestApi\interfaces\PluginInterface;
use League\Flysystem\DirectoryListing;
use League\Flysystem\FileAttributes;
use League\Flysystem\Filesystem;
use League\Flysystem\FilesystemAdapter;
use League\Flysystem\FilesystemException;
use League\Flysystem\FilesystemReader;
use League\Flysystem\PathNormalizer;
use League\Flysystem\StorageAttributes;
use League\Flysystem\UnableToCopyFile;
use League\Flysystem\UnableToCreateDirectory;
use League\Flysystem\UnableToDeleteDirectory;
use League\Flysystem\UnableToDeleteFile;
use League\Flysystem\UnableToMoveFile;
use League\Flysystem\UnableToReadFile;
use League\Flysystem\UnableToRetrieveMetadata;
use League\Flysystem\UnableToWriteFile;
use yii\base\Component;
use eluhr\flysystemRestApi\events\FileStorageEvent;

/**
 * @property FilesystemAdapter $adapter
 * @property-read  Filesystem $filesystem
 */
class FileStorage extends Component implements FileStorageComponentInterface
{
    /**
     * Abstract adapter for the filesystem
     */
    private FilesystemAdapter $_adapter;

    /**
     * Config for the filesystem
     *
     * @var array
    */
    public array $filesystemConfig = [];

    /**
     * path normalizer for the filesystem
     *
     * @var PathNormalizer|null
    */
    public ?PathNormalizer $filesystemPathNormalizer = null;

    /**
     * Used for caching the filesystem instance
     */
    private static Filesystem $_filesystem;

    /**
     * @param string $path relative path to the file
     * @param string $contents file contents as string
     * @param array $config Config options may vary between adapters
     *
     * @return bool
     */
    public function write(string $path, string $contents, array $config = []): bool
    {
        try {
            $this->trigger(FileStorageEvent::EVENT_BEFORE_WRITE, new FileStorageEvent(['data' => [
                'path' => $path,
                'contents' => $contents,
            ]]));
            $this->getFilesystem()->write($path, $contents, $config);
        } catch (FilesystemException|UnableToWriteFile $exception) {
            $this->trigger(FileStorageEvent::EVENT_EXCEPTION, new FileStorageEvent(['data' => $exception]));
            return false;
        }
        $this->trigger(FileStorageEvent::EVENT_AFTER_WRITE, new FileStorageEvent(['data' => [
            'path' => $path,
            'contents' => $contents,
        ]]));
        return true;
    }

    /**
     * @param string $path relative path to the file
     * @param resource $stream stream to write to
     * @param array $config Config options may vary between adapters
     *
     * @return bool
     */
    public function writeStream(string $path, $stream, array $config = []): bool
    {
        try {
            $this->trigger(FileStorageEvent::EVENT_BEFORE_WRITE, new FileStorageEvent(['data' => [
                'path' => $path,
                'stream' => $stream,
            ]]));
            $this->getFilesystem()->writeStream($path, $stream, $config);
        } catch (FilesystemException|UnableToWriteFile $exception) {
            $this->trigger(FileStorageEvent::EVENT_EXCEPTION, new FileStorageEvent(['data' => $exception]));
            return false;
        }

        $this->trigger(FileStorageEvent::EVENT_AFTER_WRITE, new FileStorageEvent(['data' => [
            'path' => $path,
            'stream' => $stream,
        ]]));
        return true;
    }

    /**
     * @param string $sourcePath relative path to the source file
     * @param string $destinationPath relative path to the destination file
     * @param array $config Config options may vary between adapters
     *
     * @return bool
     */
    public function move(string $sourcePath, string $destinationPath, array $config = []): bool
    {
        try {
            $this->trigger(FileStorageEvent::EVENT_BEFORE_MOVE, new FileStorageEvent(['data' => [
                'sourcePath' => $sourcePath,
                'destinationPath' => $destinationPath,
            ]]));
            $this->getFilesystem()->move($sourcePath, $destinationPath, $config);
        } catch (FilesystemException|UnableToMoveFile $exception) {
            $this->trigger(FileStorageEvent::EVENT_EXCEPTION, new FileStorageEvent(['data' => $exception]));
            return false;
        }
        $this->trigger(FileStorageEvent::EVENT_BEFORE_MOVE, new FileStorageEvent(['data' => [
            'sourcePath' => $sourcePath,
            'destinationPath' => $destinationPath,
        ]]));
        return true;
    }

    /**
     * @param string $sourcePath relative path to the source file
     * @param string $destinationPath relative path to the destination file
     * @param array $config Config options may vary between adapters
     *
     * @return bool
     */
    public function copy(string $sourcePath, string $destinationPath, array $config = []): bool
    {
        try {
            $this->trigger(FileStorageEvent::EVENT_BEFORE_COPY, new FileStorageEvent(['data' => [
                'sourcePath' => $sourcePath,
                'destinationPath' => $destinationPath,
            ]]));
            $this->getFilesystem()->copy($sourcePath, $destinationPath, $config);
        } catch (FilesystemException|UnableToCopyFile $exception) {
            $this->trigger(FileStorageEvent::EVENT_EXCEPTION, new FileStorageEvent(['data' => $exception]));
            return false;
        }
        $this->trigger(FileStorageEvent::EVENT_AFTER_COPY, new FileStorageEvent(['data' => [
            'sourcePath' => $sourcePath,
            'destinationPath' => $destinationPath,
        ]]));
        return true;
    }

    /**
     * @param string $path relative path to the file
     *
     * @return string|false
     */
    public function read(string $path)
    {
        try {
            $this->trigger(FileStorageEvent::EVENT_BEFORE_READ, new FileStorageEvent(['data' => [
                'path' => $path
            ]]));
            return $this->getFilesystem()->read($path);
        } catch (FilesystemException|UnableToReadFile $exception) {
            $this->trigger(FileStorageEvent::EVENT_EXCEPTION, new FileStorageEvent(['data' => $exception]));
        }
        return false;
    }

    /**
     * @param string $path relative path to the file
     *
     * @return resource|false
     */
    public function readStream(string $path)
    {
        try {
            $this->trigger(FileStorageEvent::EVENT_BEFORE_READ, new FileStorageEvent(['data' => [
                'path' => $path
            ]]));
            return $this->getFilesystem()->readStream($path);
        } catch (FilesystemException|UnableToReadFile $exception) {
            $this->trigger(FileStorageEvent::EVENT_EXCEPTION, new FileStorageEvent(['data' => $exception]));
        }
        return false;
    }

    /**
     * @param string $path relative path to the file
     *
     * @return string|false
     */
    public function visibility(string $path)
    {
        try {
            return $this->getFilesystem()->visibility($path);
        } catch (FilesystemException|UnableToRetrieveMetadata $exception) {
            $this->trigger(FileStorageEvent::EVENT_EXCEPTION, new FileStorageEvent(['data' => $exception]));
        }
        return false;
    }

    /**
     * @param string $path relative path to the file
     * @param string $visibility public|private
     *
     * @return bool
     * @see Visibility
     */
    public function setVisibility(string $path, string $visibility): bool
    {
        try {
            $this->getFilesystem()->setVisibility($path, $visibility);
        } catch (FilesystemException|UnableToRetrieveMetadata $exception) {
            $this->trigger(FileStorageEvent::EVENT_EXCEPTION, new FileStorageEvent(['data' => $exception]));
            return false;
        }
        return true;
    }

    /**
     * @param string $path relative path to the file
     *
     * @return bool
     */
    public function delete(string $path): bool
    {
        try {
            $this->trigger(FileStorageEvent::EVENT_BEFORE_DELETE_FILE, new FileStorageEvent(['data' => [
                'path' => $path
            ]]));
            $this->getFilesystem()->delete($path);
        } catch (FilesystemException|UnableToDeleteFile $exception) {
            $this->trigger(FileStorageEvent::EVENT_EXCEPTION, new FileStorageEvent(['data' => $exception]));
            return false;
        }
        $this->trigger(FileStorageEvent::EVENT_AFTER_DELETE_FILE, new FileStorageEvent(['data' => [
            'path' => $path
        ]]));
        return true;
    }

    /**
     * @param string $path relative path to the directory
     *
     * @return bool
     */
    public function deleteDirectory(string $path): bool
    {
        try {
            $this->trigger(FileStorageEvent::EVENT_BEFORE_DELETE_DIRECTORY, new FileStorageEvent(['data' => [
                'path' => $path
            ]]));
            $this->getFilesystem()->deleteDirectory($path);
        } catch (FilesystemException|UnableToDeleteDirectory $exception) {
            $this->trigger(FileStorageEvent::EVENT_EXCEPTION, new FileStorageEvent(['data' => $exception]));
            return false;
        }
        $this->trigger(FileStorageEvent::EVENT_AFTER_DELETE_DIRECTORY, new FileStorageEvent(['data' => [
            'path' => $path
        ]]));
        return true;
    }

    /**
     * @param string $path relative path to the directory
     *
     * @return bool
     */
    public function createDirectory(string $path): bool
    {
        try {
            $this->trigger(FileStorageEvent::EVENT_BEFORE_CREATE_DIRECTORY, new FileStorageEvent(['data' => [
                'path' => $path
            ]]));
            $this->getFilesystem()->createDirectory($path);
        } catch (FilesystemException|UnableToCreateDirectory $exception) {
            $this->trigger(FileStorageEvent::EVENT_EXCEPTION, new FileStorageEvent(['data' => $exception]));
            return false;
        }
        $this->trigger(FileStorageEvent::EVENT_AFTER_CREATE_DIRECTORY, new FileStorageEvent(['data' => [
            'path' => $path
        ]]));
        return true;
    }


    /**
     * @param string $path relative path to the directory
     * @param bool $recursive If true, list recursively the contents of the directory in a flat list.
     *
     * @return string|false
     */
    public function listContents(string $path, bool $recursive = FilesystemReader::LIST_SHALLOW): DirectoryListing|false
    {
        try {
            return $this->getFilesystem()->listContents($path, $recursive);

        } catch (FilesystemException $exception) {
            $this->trigger(FileStorageEvent::EVENT_EXCEPTION, new FileStorageEvent(['data' => $exception]));
            return false;
        }
    }

    /**
     * @param string $path
     *
     * @return bool
     */
    public function fileExists(string $path): bool
    {
        try {
            return $this->getFilesystem()->fileExists($path);
        } catch (FilesystemException|UnableToDeleteFile $exception) {
            $this->trigger(FileStorageEvent::EVENT_EXCEPTION, new FileStorageEvent(['data' => $exception]));
        }
        return false;
    }

    /**
     * File size in bytes
     *
     * @param string $path
     *
     * @return int|false
     */
    public function fileSize(string $path): bool|int
    {
        try {
            return $this->getFilesystem()->fileSize($path);
        } catch (FilesystemException|UnableToRetrieveMetadata $exception) {
            $this->trigger(FileStorageEvent::EVENT_EXCEPTION, new FileStorageEvent(['data' => $exception]));
        }
        return false;
    }

    /**
     * @param string $path
     *
     * @return string|false
     * @link https://developer.mozilla.org/en-US/docs/Web/HTTP/Basics_of_HTTP/MIME_types/Common_types
     */
    public function mimeType(string $path): bool|string
    {
        try {
            return $this->getFilesystem()->mimeType($path);
        } catch (FilesystemException|UnableToRetrieveMetadata $exception) {
            $this->trigger(FileStorageEvent::EVENT_EXCEPTION, new FileStorageEvent(['data' => $exception]));
        }
        return false;
    }

    /**
     * @param string $path
     *
     * @return int|false
     */
    public function lastModified(string $path): bool|int
    {
        try {
            return $this->getFilesystem()->lastModified($path);
        } catch (FilesystemException|UnableToRetrieveMetadata $exception) {
            $this->trigger(FileStorageEvent::EVENT_EXCEPTION, new FileStorageEvent(['data' => $exception]));
        }
        return false;
    }

    /**
     * @return FilesystemAdapter
     */
    public function getAdapter(): FilesystemAdapter
    {
        return $this->_adapter;
    }

    /**
     * @param FilesystemAdapter $adapter
     */
    public function setAdapter(FilesystemAdapter $adapter): void
    {
        $this->_adapter = $adapter;
    }

    /**
     * @return Filesystem
     */
    protected function getFilesystem(): Filesystem
    {
        // If the filesystem is not yet set, create a new instance and cache it
        if (empty(static::$_filesystem)) {
            static::$_filesystem = new Filesystem($this->getAdapter(), $this->filesystemConfig, $this->filesystemPathNormalizer);
        }
        return static::$_filesystem;
    }

    /**
     * @param StorageAttributes $storageAttributes
     *
     * @return array
     */
    public static function buildItem(StorageAttributes $storageAttributes): array
    {
        $name = basename($storageAttributes->path());
        $path = trim(dirname($storageAttributes->path()), '.');
        return [
            'id' => StorageItemHelper::generateUuid($storageAttributes->type(), $path, $name),
            'type' => $storageAttributes->type(),
            'name' => $name,
            'path' => $path,
            'size' => $storageAttributes instanceof FileAttributes ? $storageAttributes->fileSize() : null,
            'lastModified' => $storageAttributes->lastModified(),
            'extraMetadata' => $storageAttributes->extraMetadata()
        ];
    }
}

