<?php

namespace eluhr\flysystemRestApi\controllers;

use eluhr\flysystemRestApi\models\Directory;
use eluhr\flysystemRestApi\models\File;
use eluhr\flysystemRestApi\models\StorageItem;
use eluhr\flysystemRestApi\Module;
use Yii;
use yii\base\InvalidConfigException;
use yii\data\ActiveDataFilter;
use yii\data\ActiveDataProvider;
use yii\data\ArrayDataProvider;
use yii\data\DataProviderInterface;
use yii\db\Exception;
use yii\db\StaleObjectException;
use yii\rest\Controller;
use yii\web\BadRequestHttpException;
use yii\web\NotFoundHttpException;
use yii\web\UploadedFile;
use Throwable;

/**
 * @property Module $module
 */
class ApiController extends BaseController
{

    /**
     * @inheritdoc
     */
    protected function verbs(): array
    {
        $verbs = parent::verbs();
        $verbs['list'] = ['GET'];
        $verbs['list-detail'] = ['GET'];
        $verbs['create-directory'] = ['POST'];
        $verbs['delete-directory'] = ['DELETE'];
        $verbs['upload-files'] = ['POST'];
        $verbs['delete-file'] = ['DELETE'];
        $verbs['sync'] = ['GET'];
        $verbs['stream'] = ['GET'];
        $verbs['download'] = ['GET'];
        return $verbs;
    }

    /**
     * List all content details in a given path
     *
     * @param string $path
     *
     * @throws InvalidConfigException
     * @return DataProviderInterface
     */
    public function actionList(string $path): DataProviderInterface
    {

        $dataFilter = Yii::createObject(ActiveDataFilter::class);
        $dataFilter->setSearchModel(StorageItem::class);

        if ($dataFilter->load($this->request->get())) {
            $filter = $dataFilter->build();
        } else {
            $filter = [];
        }

        $query = StorageItem::find()->where(['path' => $path]);

        if (!empty($filter)) {
            $query->andWhere($filter);
        }

        return new ActiveDataProvider([
            'query' => $query,
        ]);
    }

    /**
     * Details about a file or directory
     *
     * @param string $id
     *
     * @throws NotFoundHttpException
     * @return StorageItem
     */
    public function actionListDetail(string $id): StorageItem
    {
        $model = StorageItem::findOne($id);

        if ($model === null) {
            throw new NotFoundHttpException('Resource not found.');
        }

        return $model;
    }

    /**
     * Create a new directory in the given path
     *
     * @throws BadRequestHttpException
     * @throws InvalidConfigException
     * @return array
     */
    public function actionCreateDirectory(): array
    {
        /** @var Directory $model */
        $model = Yii::createObject(Directory::class);
        $model->setFileStorage($this->getFileStorage());
        $model->setScenario(Directory::SCENARIO_CREATE);

        if ($model->load($this->request->post())) {
            return $model->getResponse();
        }

        throw new BadRequestHttpException();
    }

    /**
     * Delete a directory by an id
     *
     * @param string $id
     *
     * @throws NotFoundHttpException
     * @throws Throwable
     * @throws StaleObjectException
     * @return array
     */
    public function actionDeleteDirectory(string $id): array
    {
        $model = StorageItem::findOne($id);

        if ($model === null) {
            throw new NotFoundHttpException('Resource not found');
        }

        return [
            'success' => (bool)$model->delete(),
            'message' => ''
        ];
    }

    /**
     * Upload one or multiple files
     *
     * @throws InvalidConfigException
     * @return array
     */
    public function actionUploadFiles(): array
    {
        $fileStorage = $this->getFileStorage();

        /** @var File $model */
        $model = Yii::createObject(File::class);
        $model->setFileStorage($fileStorage);
        $model->setScenario(File::SCENARIO_UPLOAD);

        $model->basePath = $this->request->post('basePath');
        $model->files = UploadedFile::getInstances($model, 'files');

        return $model->getResponse();
    }

    /**
     * Delete a file by an id
     *
     * @param string $id
     *
     * @throws NotFoundHttpException
     * @throws Throwable
     * @throws StaleObjectException
     * @return array
     */
    public function actionDeleteFile(string $id): array
    {
        $model = StorageItem::findOne($id);

        if ($model === null) {
            throw new NotFoundHttpException('Resource not found');
        }

        return [
            'success' => $model->delete() !== false,
            'message' => ''
        ];
    }

    /**
     * @param string $path
     * @param bool $recursive
     *
     * @throws InvalidConfigException
     * @throws NotFoundHttpException
     * @throws Exception
     * @return ArrayDataProvider
     */
    public function actionSync(string $path, bool $recursive = false): ArrayDataProvider
    {
        $listing = $this->getFileStorage()->listContents($path, $recursive);

        // If the path does not exist, return false
        if ($listing === false) {
            throw new NotFoundHttpException('Resource not found');
        }

        $items = StorageItem::sync($path, $listing);

        if ($items === false) {
            throw new NotFoundHttpException('Resource not found');
        }

        return new ArrayDataProvider([
            'allModels' => $items
        ]);
    }

    public function actionStream($path)
    {
        $resource = $this->getFileStorage()->readStream($path);

        if ($resource === false) {
            throw new NotFoundHttpException('Resource not found');
        }

        return $this->response->sendStreamAsFile($resource, basename($path));
    }

    public function actionDownload($path)
    {
        return static::actionStream($path);
    }

}
