<?php
/**
 * @author: Jens Giessmann <jg@handcode.de>
 * @copyright: 2023, Jens Giessmann
 */

namespace eluhr\flysystemRestApi\interfaces;

use League\Flysystem\DirectoryListing;
use League\Flysystem\FilesystemAdapter;
use League\Flysystem\FilesystemReader;
use League\Flysystem\StorageAttributes;

interface FileStorageComponentInterface
{
    /**
     * @param string $path relative path to the file
     * @param string $contents file contents as string
     * @param array $config Config options may vary between adapters
     *
     * @return bool
     */
    public function write(string $path, string $contents, array $config = []): bool;

    /**
     * @param string $path relative path to the file
     * @param resource $stream stream to write to
     * @param array $config Config options may vary between adapters
     *
     * @return bool
     */
    public function writeStream(string $path, $stream, array $config = []): bool;

    /**
     * @param string $sourcePath relative path to the source file
     * @param string $destinationPath relative path to the destination file
     * @param array $config Config options may vary between adapters
     *
     * @return bool
     */
    public function move(string $sourcePath, string $destinationPath, array $config = []): bool;

    /**
     * @param string $sourcePath relative path to the source file
     * @param string $destinationPath relative path to the destination file
     * @param array $config Config options may vary between adapters
     *
     * @return bool
     */
    public function copy(string $sourcePath, string $destinationPath, array $config = []): bool;

    /**
     * @param string $path relative path to the file
     *
     * @return string|false
     */
    public function read(string $path);

    /**
     * @param string $path relative path to the file
     *
     * @return resource|false
     */
    public function readStream(string $path);

    /**
     * @param string $path relative path to the file
     *
     * @return string|false
     */
    public function visibility(string $path);

    /**
     * @param string $path relative path to the file
     * @param string $visibility public|private
     *
     * @return bool
     * @see Visibility
     */
    public function setVisibility(string $path, string $visibility): bool;

    /**
     * @param string $path relative path to the file
     *
     * @return bool
     */
    public function delete(string $path): bool;

    /**
     * @param string $path relative path to the directory
     *
     * @return bool
     */
    public function deleteDirectory(string $path): bool;

    /**
     * @param string $path relative path to the directory
     *
     * @return bool
     */
    public function createDirectory(string $path): bool;

    /**
     * @param string $path relative path to the directory
     * @param bool $recursive If true, list recursively the contents of the directory in a flat list.
     *
     * @return string|false
     */
    public function listContents(string $path, bool $recursive = FilesystemReader::LIST_SHALLOW): DirectoryListing|false;

    /**
     * @param string $path
     *
     * @return bool
     */
    public function fileExists(string $path): bool;

    /**
     * File size in bytes
     *
     * @param string $path
     *
     * @return int|false
     */
    public function fileSize(string $path): bool|int;

    /**
     * @param string $path
     *
     * @return string|false
     * @link https://developer.mozilla.org/en-US/docs/Web/HTTP/Basics_of_HTTP/MIME_types/Common_types
     */
    public function mimeType(string $path): bool|string;

    /**
     * @param string $path
     *
     * @return int|false
     */
    public function lastModified(string $path): bool|int;

    /**
     * @return FilesystemAdapter
     */
    public function getAdapter(): FilesystemAdapter;

    /**
     * @param FilesystemAdapter $adapter
     */
    public function setAdapter(FilesystemAdapter $adapter): void;

    /**
     * @param StorageAttributes $storageAttributes
     *
     * @return array
     */
    public static function buildItem(StorageAttributes $storageAttributes): array;

}