<?php

namespace eluhr\flysystemRestApi\models;

use DateTimeImmutable;
use eluhr\flysystemRestApi\behaviors\PluginsBehavior;
use eluhr\flysystemRestApi\components\FileStorage;
use eluhr\flysystemRestApi\helpers\StorageItemHelper;
use eluhr\flysystemRestApi\plugins\AliasPlugin;
use eluhr\flysystemRestApi\plugins\CategoryPlugin;
use eluhr\flysystemRestApi\plugins\PermissionPlugin;
use eluhr\flysystemRestApi\plugins\FileExtensionPlugin;
use eluhr\flysystemRestApi\traits\FileStorageTrait;
use League\Flysystem\DirectoryListing;
use League\Flysystem\StorageAttributes;
use yii\behaviors\TimestampBehavior;
use yii\db\ActiveRecord;
use Yii;

/**
 * @property string $id
 * @property string $path
 * @property string $name
 * @property string $type
 * @property string $last_synced_at
 */
class StorageItem extends ActiveRecord implements StorageAttributes
{

    use FileStorageTrait;

    private $_extraData = [];

    public $_label = null;

    /**
     * @inheritdoc
     */
    public static function tableName()
    {
        return '{{%storage_item}}';
    }

    /**
     * @inheritdoc
     */
    public function behaviors()
    {
        $behaviors = parent::behaviors();
        $behaviors['timestamp'] = [
            'class' => TimestampBehavior::class,
            'value' => (new DateTimeImmutable())->format('Y-m-d H:i:s')
        ];
        $behaviors['plugins'] = [
            'class' => PluginsBehavior::class,
            'plugins' => [
                PermissionPlugin::class,
                CategoryPlugin::class,
                AliasPlugin::class,
                FileExtensionPlugin::class
            ]
        ];
        return $behaviors;
    }

    /**
     * @inheritdoc
     */
    public function rules()
    {
        $rules = parent::rules();
        $rules[] = [
            [
                'path',
                'name'
            ],
            'filter',
            'filter' => function ($value) {
                if (is_string($value)) {
                    return trim($value, DIRECTORY_SEPARATOR);
                }
                return $value;
            }
        ];
        $rules[] = [
            [
                'name',
                'type',
                'last_synced_at'
            ],
            'required'
        ];
        $rules[] = [
            [
                'path',
                'name',
                'type',
            ],
            'string',
            'max' => 255
        ];
        $rules[] = [
            'last_synced_at',
            'datetime',
            'format' => 'php:Y-m-d H:i:s'
        ];
        $rules[] = [
            'type',
            'in',
            'range' => [StorageAttributes::TYPE_FILE, StorageAttributes::TYPE_DIRECTORY]
        ];
        return $rules;
    }

    /**
     * @inheritdoc
     */
    public function beforeValidate()
    {
        // update last sync time on every save
        $this->setAttribute('last_synced_at', (new DateTimeImmutable())->format('Y-m-d H:i:s'));
        return parent::beforeValidate();
    }

    /**
     * @inheritdoc
     */
    public function beforeSave($insert)
    {
        // Generate new uuid if not set
        if ($insert && empty($this->id)) {
            $this->setAttribute('id', StorageItemHelper::generateUuid($this->type, $this->path, $this->name));
        }
        return parent::beforeSave($insert);
    }

    /**
     * relative file path from  path and name
     *
     * @return string
     */
    public function getFilePath(): string
    {
        return implode(DIRECTORY_SEPARATOR, array_filter([$this->path, $this->name]));
    }

    /**
     * Sync the given path with the database
     *
     * @param string $path relative path to the root of the filesystem
     * @param DirectoryListing $listing
     *
     * @throws \yii\base\InvalidConfigException
     * @throws \yii\db\Exception
     * @return array|false List of created items or false if the path does not exist
     */
    public static function sync(string $path, DirectoryListing $listing): array
    {
        $transaction = static::getDb()->beginTransaction();
        static::deleteAll(['path' => $path]);
        $items = [];
        /** @var StorageAttributes $item */
        foreach ($listing as $item) {
            $name = basename($item->path());
            $path = trim(dirname($item->path()), '.');
            if (!StorageItemHelper::upsertItemInDb($item->type(), $path, $name)) {
                $transaction->rollBack();
                return false;
            } else {
                $items[] = static::getFileStorage()::buildItem($item);
            }
        }

        $transaction->commit();

        return $items;
    }

    /**
     * @inheritdoc
     */
    public function fields()
    {
        return [
            'uuid' => 'id',
            'path' => function (self $model) {
                return $model->path();
            },
            'name' => function (self $model) {
                return $model->getLabel();
            },
            'type' => function (self $model) {
                return $model->type();
            },
            'size' => function (self $model) {
                return $model->type === StorageAttributes::TYPE_FILE ? $model->getFileStorage()->fileSize($model->getFilePath()) : null;
            },
            'lastModified' => function (self $model) {
                return $model->getFileStorage()->lastModified($model->getFilePath());
            },
            'extraMetadata' => function (self $model) {
                return $model->extraMetadata();
            }
        ];
    }

    public function offsetUnset(mixed $offset): void
    {
        parent::offsetUnset($offset);
    }

    public function path(): string
    {
        return $this->path;
    }

    public function type(): string
    {
        return $this->type;
    }

    public function visibility(): ?string
    {
        return $this->getFileStorage()->visibility($this->getFilePath());

    }

    public function lastModified(): ?int
    {
        return $this->getFileStorage()->lastModified($this->getFilePath());
    }

    public static function fromArray(array $attributes): StorageAttributes
    {
        $model = Yii::createObject(static::class);
        $model->setAttributes($attributes);
        return $model;
    }

    public function isFile(): bool
    {
        return $this->type() === StorageAttributes::TYPE_FILE;
    }

    public function isDir(): bool
    {
        return $this->type() === StorageAttributes::TYPE_DIRECTORY;
    }

    public function withPath(string $path): StorageAttributes
    {
        $model = clone $this;
        $model->path = $path;
        return $model;
    }

    public function extraMetadata(): array
    {
        return $this->_extraData;
    }

    public function jsonSerialize(): mixed
    {
        return $this->fields();
    }

    /**
     * @inheritdoc
    */
    public function beforeDelete()
    {
        if ($this->isDir()) {
            $deletedFile = $this->getFileStorage()->deleteDirectory($this->getFilePath());
        } else {
            $deletedFile = $this->getFileStorage()->delete($this->getFilePath());
        }

        // TODO: What about files in (sub)folders?

        return $deletedFile && parent::beforeDelete();
    }

    public function addExtraMetaData(string $name, $data)
    {
        return $this->_extraData[$name] = $data;
    }

    /**
     * @return string
     */
    public function getLabel(): string
    {
        return $this->_label ?: $this->name;
    }

    /**
     * @param string $label
     */
    public function setLabel(string $label): void
    {
        $this->_label = $label;
    }
}
